#!/Library/Frameworks/Python.framework/Versions/3.6/bin/python3.6

# Copyright (C) 2018 Matthew L. Schwartz - all rights reserved
# Not for redistribution

import SapTrap_Classes as STC
import SapTrap_Functions as STF

import tkinter as tk
from tkinter import messagebox
from functools import partial   # Allows button commands to pass an argument to called function
import tkinter.font as tkf


# Base Window Setup
root = tk.Tk()
root.configure(bg='grey', border=10)
root.title('SapTrap Builder')
display_line_tags = 15


#dialog for 'about program'
def aboutSapTrap():
    messagebox.showinfo('SapTrap Builder',
                        'Version : 4.2.1\nCopyright (C) 2018 Matthew L. Schwartz'
                        '\n\nPython 3.6 - Copyright (C) 2001-2016 Python Software Foundation; All Rights Reserved'
                        '\nTcl/Tk Copyright (C) This software is copyrighted by the Regents of the '
                        '\nUniversity of California, Sun Microsystems, Inc., Scriptics Corporation, and other parties. ')

root.createcommand('tkAboutDialog', aboutSapTrap)



# Steps added to program_control library will automatically be incorporated into main display
# if adding steps or changing step names, need to change name in program_controller function (line 103)
# Program controller directory should be the only place where coordination is needed

dict_program_control = {'1':'Sequence Entry', '2':'Select Translation Frame', '3':'Select Tag Type',
                   '4':'Set Insert Site', '5':'Select gRNA', '6':'Mutate Repair Template',
                   '7':'Review Design', '8':'Get Reagent Sequences'}

#dict_program_control = {'1':'Sequence Entry', '2':'Select Translation Frame', '3':'Mutate Repair Template',
#                        '4':'Select gRNA', '5':'Set Insert Site', '6':'Select Tag Type',
#                        '7':'Review Design', '8':'Get Reagent Sequences'}

keys_program_section_control = list(dict_program_control.keys())
bases = ['a', 'c', 'g', 't', '-']
amino_acids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M',
               'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y', '*']
additional_controls = ['z', 'r']  #'z' is set to master reset (back to WT), r to local (session start) reset of mutator
#temporary_directionals = ['<[>', '<]>', '<{>', '<}>', '<,>', '<.>', 'u', 'x'] # temp to get windows keys working
directional = [ '<Left>', '<Right>', '<Up>', '<Down>' ,
                '<Shift-Left>', '<Shift-Right>', '<Shift-Up>', '<Shift-Down>' ]
#temporary_directionals_bindings = {'[':'left', ']':'right', '{':'LEFT', '}':'RIGHT',    # temp for windows
#                                   'u':'up', 'x':'down', ',':'UP', '.':'DOWN'}
tab_controls = ['<Tab>', '<Shift-Tab>']

def get_user_response(event):
    """ Triggered by key and button presses, sets user_response variable based on event. """
    text = ''
    if event.keysym in ['Up', 'Down', 'Left', 'Right']:
        if event.state in [96, 262152]:
            text = event.keysym.lower()
        elif event.state in [97, 262153]:
            text = event.keysym.upper()
#    elif event.char in temporary_directionals_bindings.keys():  # temp for windows
#        text = temporary_directionals_bindings[event.char]
    else:
        text = event.char
    user_response.set(text)

    return 0

# Bind specific keys to get_user_response for user interaction
for list in [keys_program_section_control, amino_acids, bases, additional_controls, directional,
             tab_controls]:
    for key in list:
        root.bind(key, get_user_response)


def set_user_response(value):
    """ This function allows buttons to set the user_response variable. Used for button based flow control."""
    user_response.set(value)

    return 0

def step_forward():
    steps = {'1':'2', '2':'3', '3':'4', '4':'5', '5':'6', '6':'7', '7':'8'}

    if str(progress_tracker.get()) in steps.keys():
        string_next_step = steps[str(progress_tracker.get())]
        next_step = dict_program_control[string_next_step]
        return set_user_response(next_step)
    else:
        return 0

def step_backward():
    steps = {'8': '7', '7': '6', '6': '5', '5': '4', '4': '3', '3': '2', '2': '1'}

    if str(progress_tracker.get()) in steps.keys():
        string_next_step = steps[str(progress_tracker.get())]
        next_step = dict_program_control[string_next_step]
        return set_user_response(next_step)
    else:
        return 0

def highlight_active_step_button():
    active_step = progress_tracker.get()
    for i in range(8):
        if i == active_step - 1:
            progress_buttons[i].configure(font=('TkDefaultFont', '13', 'bold'))
        else:
            progress_buttons[i].configure(font=('TkDefaultFont', '13'))

    return 0

def tab_selector(event):
    if event.keysym == 'Tab':
        if event.state in [0, 8]:
            step_forward()
            return 'break'
        else:
            return step_backward()


def do_nothing():
    """ I use this function to inactivate the enter_sequence button outside of step 1 """
    return 0


def format_sequence(event):
    if progress_tracker.get() == 1:
        sequence = main_display.get('1.0', tk.END)
        try:
            length = len(main_display.get(tk.SEL_FIRST, tk.SEL_LAST))
        except:
            length = 100

        sequence = STF.strip_DNA_sequence(sequence, length)
        main_display.delete('1.0', tk.END)
        main_display.replace('1.0', tk.END, sequence)

    return 0


def change_selection_case(event):
    if progress_tracker.get() == 1:
        try:
            sequence = main_display.get(tk.SEL_FIRST, tk.SEL_LAST)
            change = event.char
            sequence = STF.change_DNA_case(sequence, change)
            main_display.replace(tk.SEL_FIRST, tk.SEL_LAST, sequence)
        except:
            return 0

    return 0

def quit(event):
    root.destroy()

root.bind('<Command-s>', format_sequence)
root.bind('<Command-u>', change_selection_case)
root.bind('<Command-l>', change_selection_case)
root.bind('<Command-d>', quit)


def final_entry_validation():
    """ Checks data states and returns False if any data needed for steps 7 or 8 are missing.
        Make sure that none of the required fields / values are set to None or inappropriate values
        before determining homology arms for final preview or final calculation of reagents """
    if site.tag_type is None:
        try:
            site.tag_type = tag_keys[tag_type.get()]
        except:
            messagebox.showerror('Tag type', 'Please select a tag type (step 3)')
            return False
    elif site.tag_type is not None:
        site.tag_type = tag_keys[tag_type.get()]
    if site.gRNA == None:
        messagebox.showerror('guide RNA', 'Please select a guide RNA (step 5)')
        return False
    for user_input in [hUS_length_entry, hDS_length_entry, min_length_entry]:
        try:
            int(user_input.get('1.0', tk.END))
        except ValueError:
            messagebox.showerror('hUS target length', 'hUS target length field must be set to an integer')
            return False
    site.hUS_target_length = int(hUS_length_entry.get('1.0', tk.END))
    site.hDS_target_length = int(hDS_length_entry.get('1.0', tk.END))
    site.homology_arm_minimum_length = int(min_length_entry.get('1.0', tk.END))
    site.jobname = jobname.get()
    site.sap_option = sap_selection.get()


    return True


def scrollset():
    """Functionalizing the scrollset allows you to reset the scrollbars for new or changed text"""
    scrollX = tk.Scrollbar(root, orient=tk.HORIZONTAL, command=main_display.xview)
    scrollX.grid(row=17, column=1, columnspan=4, sticky=tk.NE + tk.NW)
    main_display['xscrollcommand'] = scrollX.set  # This allows marker to travel entire bar, and to stick where left
    scrollY = tk.Scrollbar(root, orient=tk.VERTICAL, command=main_display.yview)
    scrollY.grid(row=2, column=5, rowspan=15, sticky=tk.NW + tk.SW)
    main_display['yscrollcommand'] = scrollY.set


def update_display(display_text):
    """ Input is library with 'sequences' for main_display and 'labels' for label_display.
        Changes text in main_display and label_display. """
    # now need to loop through creating and then applying the line tags,
    # then add text to include line flags in output from display calls
    main_display.configure(state='normal')
    main_display.delete('1.0', tk.END)
    main_display.replace('1.0', tk.END, display_text['sequences'])
    for i in display_text['flags']:
        main_display.tag_add(str(i), str(i) + '.0', str(i) + '.end')
        main_display.tag_config(str(i), foreground='red')
    main_display.configure(state='disabled')
    label_display.configure(text=display_text['labels'])
    scrollset()

def section_loop(section_function):
    """ This loop works with the individual section functions.
        Updates the main_display box and waits for user_resopnse to change
        Section functions are designed to implement previous user_response, then display new user choice
        Loop exits when new program_control input is placed in user_response. """
    input_test = True
    while input_test is True:
        update_display(display.format_display_strings())
        root.wait_variable(user_response)
        if user_response.get() in dict_program_control.values():
            section_loop_exit.set(True)
            input_test = False
        else:
            input_test = section_function()

    return 0


def incrementer(initial_value, maximum, minimum=0, increment=1, circular=False):
    """ Increments through a range of integers(minimum, maximum) by increment(default = 1)
        If circular = 'No', will return minimum or maximum when asked to increment beyond range
        If circular = 'Yes', minimum - 1 = maximum, maximum + 1 = minimum
        Returns incremented value """

    initial_sum = initial_value + increment
    if initial_sum in range(minimum, maximum + 1):
        return initial_sum
    elif circular == True:
        true_increment = ((increment - (maximum  + 1 - initial_value)) % (maximum + 1 - minimum))
        return minimum + true_increment
    elif circular == False:
        if initial_sum > maximum:
            return maximum
        elif initial_sum < minimum:
            return minimum


Cas9 = STC.GuideSpecification('NGG', -20, -1, -3, 'RIGHT', 'Cas9')
Cpf1 = STC.GuideSpecification('AAAN', 4, 23, 25, 'LEFT', 'Cpf1')

def enter_sequence_setup():
    ''' Clears the main_display and opens access for user input. Activates enter_button. '''
    reset_site = True

    if progress_tracker.get() != 1:
        reset_site = messagebox.askokcancel('Delete Current Site?',
                                            'Returning to Sequence Entry will delete current site!')
        if reset_site:
            site.reset('CAT', Cas9)
            display.reset(site)
        else:
            label_text = program_instructions_labels(dict_program_control[str(progress_tracker.get())])
            main_label.configure(text=label_text)
            return 0
    label_display.configure(text='')
    main_display.configure(state='normal')
    #enter_button.configure(command=enter_sequence)
    main_display.delete('1.0', tk.END)
    progress_tracker.set(1)
    highlight_active_step_button()

    return 0


def enter_sequence():
    """ Reads in sequence from main_display.
        Display message if sequence does not conform to program requirements.
        Resets Edit_site object to input values if user input is accepted.
        Directs program to step 2. """
    import_sequence = STF.format_sequence(main_display.get('1.0', tk.END))
    guide = Cas9
    if import_sequence['change_nuclease']:
        guide = Cpf1
    if import_sequence['sequence'] == 'Entry_Error':
        messagebox.showerror('Input Error', 'Entry sequence can only contain only:\n1. DNA base characters '
                                            '\n2. New line characters (returns)\n3. A single, '
                                            'optional / to mark the target insertion site.'
                                            '\n4. The characters < and > to mark the start and end of a '
                                            'large intended deletion.')
        return False
    elif import_sequence['sequence'] == '':
        messagebox.showerror('Input Error', 'Please enter a site.')
        return False
    else:
        main_display.delete('1.0', tk.END)
        site.reset(import_sequence['sequence'], guide, insert_site=import_sequence['insertion_site'])
        #enter_button.configure(command=do_nothing)
        user_response.set(dict_program_control['2'])
    if import_sequence['deletion'] == True:
        site.incorporate_large_deletion(import_sequence['deletion_region'])
        display.add_display_strings('mutated_sequence')
    return True


def select_translation_frame_setup():
    """ Updates display_strings to contian necessary strings.
        scrolls to insert_site. """
    display.remove_display_strings('repair_template')
    display.add_display_strings('base_translation', 'sequence')
    update_display(display.format_display_strings())
    main_display.see(site.feature_locator('insert_site'))
    progress_tracker.set(2)
    highlight_active_step_button()

    return 0


def select_translation_frame():
    """ reads user_response entries of left and right - changes reading frame in increments of +- 1 """
    direction = user_response.get()
    if direction in ['right', 'up']:
        site.change_frame(1)
    elif direction in ['left', 'down']:
        site.change_frame(2)
    site.base_translation = site.translate(site.sequence)
    site.mutant_translation = site.translate(site.mutated_sequence)

    return True


def select_tag_type_setup():
    """ Currently tag_type selection is handled entirely by an OptionMenu
        Empty functions are placeholders for later addition of a dialog to get this input. """
    progress_tracker.set(3)
    highlight_active_step_button()
    return 0


def select_tag_type():
    """ Input is up or down arrow key
        Sets option menu to previous or next option from list """
    current_tag_type = tag_type.get()
    max_tag_index = len(tag_types) - 1
    try:
        current_tag_index = tag_types.index(current_tag_type)
    except:
        current_tag_index = -1
    response = user_response.get()
    if response == 'up':
        new_tag_index = incrementer(current_tag_index, max_tag_index, increment=-1, circular=True)
        tag_type.set(tag_types[new_tag_index])
    elif response == 'down':
        new_tag_index = incrementer(current_tag_index, max_tag_index, increment=1, circular=True)
        tag_type.set(tag_types[new_tag_index])

    return True


def set_insert_site_setup():
    """ updates display_strings to show relevant information.
        scrolls to initial insert_site. """
    site.show_insert_site = True # initializes to false which suppresses '-INSERT-' text from display before this step
    display.remove_display_strings('repair_template')
    display.add_display_strings('base_translation', 'sequence')
    main_display.see(site.feature_locator('insert_site'))
    progress_tracker.set(4)
    highlight_active_step_button()

    return 0


def set_insert_site():
    """ reads user_response for right, left, RIGHT, LEFT, 'c', and 'r'
        Directionals move the insert_site, r resets to original, and c moves to cutsite of gRNA, if not None """
    response = user_response.get()
    if response == 'right':
        site.insert_site += 1
    elif response == 'left':
        site.insert_site -= 1
    elif response == 'RIGHT':
        site.insert_site += 9
    elif response == 'LEFT':
        site.insert_site -= 9
    elif response == 'c':
        if site.gRNA is not None:
            site.insert_site = site.gRNA.cutsite
    elif response == 'r':
        site.insert_site = site.original_insert_site
    if site.gRNA is not None:
        site.gRNA.distance = abs(site.gRNA.cutsite - site.insert_site)
        site.PAM_string = site.gRNA.aligned_pam()
    main_display.see(site.feature_locator('insert_site') + '+10c')
    main_display.see(site.feature_locator('insert_site') + '-5c')

    return True


def select_gRNA_setup():
    """ Adjusts display_strings.
        Orders gRNA list based on proximity to set insert site, and if no gRNA selected, selects closest cutting gRNA
        """
    selection = 1
    for i in range(1, 1 + len(site.gRNA_list)):
        site.gRNA_list[i].distance = abs(site.gRNA_list[i].cutsite - site.insert_site)
        if site.gRNA_list[i].distance < site.gRNA_list[selection].distance:
            selection = i
    if site.gRNA is None:
        site.set_gRNA(selection)
    display.remove_display_strings('repair_template')
    display.add_display_strings('base_translation', 'sequence', 'gRNA', 'PAM')
    main_display.see(site.feature_locator('guide'))
    progress_tracker.set(5)
    highlight_active_step_button()

    return 0


def select_gRNA():
    """ Reads user_response variable for left or right - loops through guides in guide_list"""
    max_guide = len(site.gRNA_list)
    if max_guide is None:       # Need to test this that loop exits upon no guides found
        return False
    response = user_response.get()
    if response == 'right':
        site.set_gRNA(incrementer(site.gRNA.index, max_guide, minimum=1, circular=True))
    if response == 'left':
        site.set_gRNA(incrementer(site.gRNA.index, max_guide, minimum=1, increment=-1, circular=True))
    main_display.see(site.feature_locator('guide') + '+20c')
    main_display.see(site.feature_locator('guide') + '-20c')
    return True
    # update_display(site.format_display_strings(site.get_display_strings()))
    # need to fitx guidestring and pamstring calls


def mutatate_repair_template_setup():
    """ Stores snapshot of mutated_sequence attribute in temporary_sequence_hold, updates display_strings attribute
        Sets mutation_site pointer to gRNA if available"""
    site.temporary_sequence_hold = site.mutated_sequence # Snapshot of mutated_sequence on entry for reset function
    display.remove_display_strings('repair_template')
    display.add_display_strings('base_translation', 'translation_differences', 'mutant_translation', 'pointer',
                                'mutated_sequence', 'DNA_differences', 'sequence', 'gRNA', 'PAM', 'SapI_sites')
    if site.gRNA is not None:
        site.mutation_site = site.gRNA.cutsite
    else:
        site.mutation_site = site.insert_site
    site.set_pointer(site.mutation_site)
    #site.generate_SapI_string()
    main_display.see(site.feature_locator('mutation'))
    progress_tracker.set(6)
    highlight_active_step_button()

    return 0


def mutate_repair_template():
    """ Reads current value of user_response, and either alters the Edit_site objects mutated_sequence attribute
        or moves the pointer (alters the Edit_site objects mutation_site attribute"""
    # Get user response and store mutation site variables locally
    response = user_response.get()
    mutation_site = site.mutation_site
    mutated_sequence = site.mutated_sequence
    # Reset current work, or reset all work; incorporate_mutation called on silent mutation to reset translations
    if response == 'r':
        site.mutated_sequence = site.temporary_sequence_hold
        site.incorporate_mutation(mutation_site, site.mutated_sequence[mutation_site])  # silent, resets translations
    elif response == 'z':
        site.mutated_sequence = site.sequence
        site.incorporate_mutation(mutation_site, site.mutated_sequence[mutation_site])  # silent, resets translations
    # Left and Right move the pointer to new positions along the sequence
    if response == 'right':
        site.mutation_site = incrementer(mutation_site, len(site.sequence) - 1)
    elif response == 'left':
        site.mutation_site = incrementer(mutation_site, len(site.sequence), increment=-1)
    elif response == 'RIGHT':
        site.mutation_site = incrementer(mutation_site, len(site.sequence) - 1, increment=9)
    elif response == 'LEFT':
        site.mutation_site = incrementer(mutation_site, len(site.sequence) - 1, increment=-9)
    if site.mutated_sequence[mutation_site] == '-':
        if response in bases:
            if site.sequence[mutation_site].isupper():
                site.incorporate_mutation(mutation_site, response.upper())
            elif site.sequence[mutation_site].islower():
                site.incorporate_mutation(mutation_site, response.lower())
    elif site.mutated_sequence[mutation_site].islower():
        if response in bases:
            site.incorporate_mutation(mutation_site, response.lower())
    # Handle mutations when active base is coding
    elif site.mutated_sequence[mutation_site].isupper():
        if response in bases:
            site.incorporate_mutation(mutation_site, response.upper())
        elif response == '-':
            site.incorporate_mutation(mutation_site, '-')
        elif site.get_current_codon(mutation_site) != {}:
            current_codon_information = site.get_current_codon(mutation_site)
            current_codon_positions = current_codon_information['positions_list']
            current_codon = current_codon_information['codon']
            current_amino_acid_index = amino_acids.index(current_codon_information['amino_acid'])
            silent_mutations = STF.synonymous_codons(current_codon_information['amino_acid'])
            current_codon_index = silent_mutations.index(current_codon)
            new_codon = ''
            new_amino_acid = ''
            if response == 'up':
                new_codon = silent_mutations[incrementer(current_codon_index, len(silent_mutations) - 1, increment=-1)]
            elif response == 'down':
                new_codon = silent_mutations[incrementer(current_codon_index, len(silent_mutations) - 1)]
            elif response == 'UP':
                new_amino_acid = amino_acids[incrementer(current_amino_acid_index, len(amino_acids) - 1,
                                             circular=True)]
            elif response == 'DOWN':
                new_amino_acid = amino_acids[incrementer(current_amino_acid_index, len(amino_acids) - 1, increment=-1,
                                             circular=True)]
            elif response in amino_acids:
                new_amino_acid = response
            # Apply mutations to the mutated_sequence string
            if new_amino_acid != '':
                new_codon = STF.synonymous_codons(new_amino_acid)[0]
            if new_codon != '':
                for i in range(3):
                    if current_codon[i] == new_codon[i]:
                        continue
                    else:
                        site.incorporate_mutation(current_codon_positions[i], new_codon[i])
    # Update relevant display_string sources for next iteration
    if mutation_site != site.mutation_site:
        site.set_pointer(site.mutation_site)
    if mutated_sequence != site.mutated_sequence:
        site.mutant_translation = site.translate(site.mutated_sequence)
        site.DNA_difference_string = site.difference_string('DNA')
        site.translation_difference_string = site.difference_string('Translation')
        site.generate_SapI_string()
    main_display.see(site.feature_locator('mutation') + '+10c')
    main_display.see(site.feature_locator('mutation') + '-10c')
    return True


def review_design_setup():
    """ Changes Edit_site object attributes: homology_arm_flags, display_strings """
    site.homology_arm_flags = []
    site.DNA_difference_string = site.difference_string('DNA-Final')
    site.site_review()
    site.review_flags()
    display.add_display_strings('base_translation', 'translation_differences', 'mutant_translation',
                                'DNA_differences', 'sequence', 'gRNA', 'PAM', 'repair_template',
                                'SapI_sites', 'homology_arm_flags')
    display.remove_display_strings('pointer', 'mutated_sequence')
    progress_tracker.set(7)
    highlight_active_step_button()


def review_design():
    """ Displays summary of design with current user options.
        Displays flags for Homology arm design, general site, and mutagenesis """
    update_display(display.format_display_strings())
    #main_display.configure(state='normal')
    #main_display.insert(tk.END, site.review_flags())
    #main_display.configure(state='disabled')
    main_display.see(site.feature_locator('insert_site'))

    return 0


def get_reagent_sequences_setup():
    progress_tracker.set(8)
    highlight_active_step_button()

    return 0


def get_reagent_sequences():
    """ Final step - Generates output sequences and prints them to main_display.
        sets main_display to normal behavior to allow user to copy text. """
    site.get_homology_arms()
    site.review_flags()
    update_display(site.print_oligos(output_type.get()))
    main_display.configure(state='normal')
    main_display.see('1.0')
    #main_display.replace('1.0', tk.END, site.print_oligos(output_type.get()))
    # label_display.configure(text='')
    progress_tracker.set(8)

def program_instructions_labels(identifier):
    """ Input = step identifier, returns instructions string. """
    label = identifier + '\n'
    if identifier == 'Sequence Entry':
        label += '1. Enter DNA sequence (CAPS for coding, lowercase for non-coding sequence).' \
                 '\tcommand+s = strip non-DNA characters' \
                 '\n2. (Optional) Mark insert site with / , or a region to delete with <....>.' \
                 '\t\tcommand+u = selection to UPPER CASE' \
                 '\n3. Press Tab, \'>\', or Step button to advance' \
                 '\t\t\t\t\tcommand+l = selection to lower case'
    elif identifier == 'Select Translation Frame':
        label += 'Left, Right arrow keys = shift translation frame +- 1.'
    elif identifier == 'Select Tag Type':
        label += 'Select tag type from dropdown list.\nUse Up and Down arrow keys to scroll through options.'
    elif identifier == 'Set Insert Site':
        label += 'Left, Right arrow keys = move insert site +- 1 base. (CAPS = in frame)' \
                 '\nShift+Left, Shift+Right arrow keys = move insert site +- 9 bases.' \
                 '\nr = reset to original position \tc = set insert site to cut site of selected gRNA'
    elif identifier == 'Select gRNA':
        label += 'Left, Right arrow keys = select next nearest gRNA candidate.'
    elif identifier == 'Mutate Repair Template':
        label += 'Up, Down = synonymous codons\t\t\ta, c, g, t = change base\tr = session reset' \
                 '\nShift+Up, Shift+Down = non-synonymous codons\t- = delete base\t\tz = reset to wild type ' \
                 '\nLeft, Right = Navigate (+Shift for 9 base steps)\tA, C, D, ... = change amino acid\t'
    elif identifier == 'Review Design':
        label += 'Repair Template trimmed to specified length.'
    elif identifier == 'Get Reagent Sequences':
        label += ''

    return label

def program_controller(identifier):
    """ Input is a string corresponding to a value in the dict_program_control dictionary
        Button presses set user_response to one of the values in the dict_program_control dictionary
        in mainloop, when user_response is set to one of these values, this program is called
        Each branch of the program can perform setup activities in the _setup method,
        then enter a user interactive loop"""
    root.focus_force()
    if progress_tracker.get() == 1:
        if not enter_sequence():
            return 0
    #if str(site.sequence) == 'CAT':
        #messagebox.showinfo('Error', 'Enter sequence and press Enter Sequence button before proceeding.')
        #return 0
    label_text = program_instructions_labels(identifier)
    main_label.configure(text=label_text)
    if identifier == 'Sequence Entry':
        enter_sequence_setup()
    elif identifier == 'Select Translation Frame':
        select_translation_frame_setup()
        section_loop(select_translation_frame)
    elif identifier == 'Select Tag Type':
        select_tag_type_setup()
        section_loop(select_tag_type)
    elif identifier == 'Set Insert Site':
        set_insert_site_setup()
        section_loop(set_insert_site)
    elif identifier == 'Select gRNA':
        if site.gRNA_list != {}:
            select_gRNA_setup()
            section_loop(select_gRNA)
        else:
            messagebox.showinfo('No guide RNA Candidates', 'This sequence contains no guide RNA candidates!')
            main_label.configure(text='Select action')
    elif identifier == 'Mutate Repair Template':
        mutatate_repair_template_setup()
        section_loop(mutate_repair_template)
    elif identifier == 'Review Design':
        if final_entry_validation():    # Ensures all required user entry / selections have been made
            review_design_setup()
            review_design()
    elif identifier == 'Get Reagent Sequences':
        if final_entry_validation():    # Ensures all required user entry / selections have been made
            get_reagent_sequences_setup()
            get_reagent_sequences()
    elif identifier == 'Quit':
        root.destroy()
    else:
        messagebox.showerror('Error', 'Program controller received out of bounds identifier.')
    return 0


# Variables used in main GUI interface
program_control_method_selector = [enter_sequence, set_insert_site, select_tag_type, select_translation_frame,
                                   select_gRNA, mutate_repair_template, get_reagent_sequences]
# tk variables controlled by user input
user_response = tk.StringVar()
section_loop_exit = tk.BooleanVar()
jobname = tk.StringVar(root)
sap_selection = tk.StringVar(root)
output_type = tk.StringVar(root)
tag_type = tk.StringVar(root)
progress_tracker = tk.IntVar(root)
# tk variable setup - OptionMenu choices, initial value setting
sap_options = {'SapTrap', 'Generic'}
sap_selection.set('SapTrap')
output_options = ['Oligos + Annotated', 'Oligos', 'Synthetic DNA', 'Annotated Synthetic DNA',  'Alignment']
output_type.set('Oligos + Annotated')
tag_types = ['N-tag', 'C-tag', 'Internal - with Linkers', 'Internal, no Linkers', 'Native Intron', 'Syntron']
tag_keys = {'N-tag': 'N', 'C-tag': 'C', 'Internal - with Linkers': 'L', 'Internal, no Linkers': 'X',
            'Native Intron': 'I', 'Syntron': 'S'}
tag_type.set(dict_program_control[str(program_control_method_selector.index(select_tag_type) + 1)]) # sets label
progress_tracker.set(1)

# main_width is outer window size; sub_width is size of text boxes inside window
main_width = int(0.9 * root.winfo_screenwidth())
main_height = int(0.5 * root.winfo_screenheight())
font = tkf.Font(family='Courier', size='14')
label_display_width = 30 * int(font.measure('m'))   # display width of label_display in pixels
main_display_width = int((main_width - label_display_width - 40) /
                         font.measure('m'))         # max width of main_display in characters
# sub_width = int(0.9 * main_width)
# sub_height = int(0.9 * main_height)
root.geometry('{0}x{1}'.format(main_width, main_height))

for x in range(5):
    root.columnconfigure(x, weight=1)
for y in range(30):
    root.rowconfigure(y, weight=1)


# Settings Frame - contains default parameters user can edit to adjust general program output
settings = tk.Frame(root)
settings.grid(row=0, column=0, columnspan=3, sticky=tk.W)
progress_bar = tk.Frame(root)
progress_bar.grid(row=0, column=3, columnspan=2, sticky=tk.E)
main_label = tk.Label(root, bg='grey', height=5, justify=tk.LEFT, anchor=tk.N,
                      text=program_instructions_labels('Sequence Entry'))
main_label.grid(row=18, column=1, columnspan=4, sticky=tk.NW)
main_display = tk.Text(root, height=15, state='normal', wrap='none', font=('Courier', '14'))
main_display.grid(row=2, column=1, columnspan=4, rowspan=15, sticky=tk.W+tk.E+tk.N+tk.S)
label_display = tk.Label(root, width=30, height=15, anchor=tk.NE, bg='grey', justify=tk.RIGHT, font=('Courier', '14'))
label_display.grid(row=2, column=0, rowspan=15, sticky=tk.E+tk.N+tk.S+tk.W)
directional_control = tk.Frame(root)
directional_control.grid(row=1, column=4, columnspan=1, sticky=tk.NE)


# Header labels for column sets 1 and 2
job_info_label = tk.Label(settings, text='Job Information')
job_info_label.grid(row=0, column=0, columnspan=2)
column_spacer = tk.Label(settings, text='    ')
column_spacer.grid(row=0, column=2)
homology_arm_label = tk.Label(settings, text='Homology Arm Setup')
homology_arm_label.grid(row=0, column=3, columnspan=2)
# Labels for column set 1
jobname_label = tk.Label(settings, text='Job Name:')
jobname_label.grid(row=1, column=0)
sap_selector_label = tk.Label(settings, text='  Type:')
sap_selector_label.grid(row=2, column=0)
output_type_label = tk.Label(settings, text=' Output:')
output_type_label.grid(row=3, column=0)
# User entry widgets for column set 1
jobname_entry = tk.Entry(settings, width=20, font=('Courier', '14'), textvariable=jobname)
jobname_entry.grid(row=1, column=1)
sap_selector = tk.OptionMenu(settings, sap_selection, *sap_options)
sap_selector.grid(row=2, column=1, sticky=tk.E + tk.W)
output_type_selector = tk.OptionMenu(settings, output_type, *output_options)
output_type_selector.grid(row=3, column=1, sticky=tk.E + tk.W)
# Labels for column set 2
hUS_length_label = tk.Label(settings, height = 1, text='hUS target length: ')
hUS_length_label.grid(row=1, column=3)
hDS_length_label = tk.Label(settings, height = 1, text='hDS target length: ')
hDS_length_label.grid(row=2, column=3)
min_length_label = tk.Label(settings, height=1, text='Min true length:')
min_length_label.grid(row=3, column=3)
# User entry widgets for column set 2
hUS_length_entry = tk.Text(settings, height=1, width=3, font=('Courier', '14'))
hUS_length_entry.grid(row=1, column=4)
hUS_length_entry.insert(tk.END, '57')
hDS_length_entry = tk.Text(settings, height=1, width=3, font=('Courier', '14'))
hDS_length_entry.grid(row=2, column=4)
hDS_length_entry.insert(tk.END, '57')
min_length_entry = tk.Text(settings, height=1, width=3, font=('Courier', '14'))
min_length_entry.grid(row=3, column=4)
min_length_entry.insert(tk.END, '32')

# Progress Menu - Contains Button widgets outlining the suggested program flow

progress_labels = []
# Generates numeric labels in front of buttons specifying suggested step order
# This has been commented out temporarily to leave the labels unnumbered
#for key in dict_program_control.keys():
#    index = int(key) - 1
#    row = index % 4
#    column = 2 * int(index / 4)
#    progress_labels.append(tk.Label(progress_bar, text=key))
#    progress_labels[index].grid(row=row, column=column)
# Generates buttons and an option menu corresponding to thd individual tasks of the workflow
progress_buttons = []
for key in dict_program_control.keys():
    index = int(key) - 1
    row = index % 4
    column = 1 + 2 * int(index / 4)
    if dict_program_control[key] == 'Select Tag Type':
        progress_buttons.append(tk.OptionMenu(progress_bar, tag_type, *tag_types))
    else:
        progress_buttons.append(tk.Button(progress_bar, text=dict_program_control[key], width=20,
                                          command=partial(set_user_response, dict_program_control[key])))
    progress_buttons[index].grid(row=row, column=column, sticky=tk.E+tk.W)

# main_display - setup for main frame, main_display window, and user enter_button

#enter_button = tk.Button(root, text='Enter Sequence', command=enter_sequence)
#enter_button.grid(row=4, column=1, sticky=tk.E)
# Directional control Frame - houses forward and back buttons under Progress Menu

advance_button = tk.Button(directional_control, text='>', font=('Courier', '20', 'bold'),
                           command=step_forward)
advance_button.grid(row=0, column=1, sticky=tk.E)
regress_button = tk.Button(directional_control, text='<', font=('Courier', '20', 'bold'),
                           command=step_backward)
regress_button.grid(row=0, column=0, sticky=tk.E)


for tab in ['<Tab>', '<Shift-Tab>']:
    root.bind(tab, tab_selector)
    main_display.bind(tab, tab_selector)

# Main Program. Initializes the Edit_site object with minimal sequence 'CAT' and sets program to entry box
# Only a signle Edit_site object is ever created by program; values reset by entry in step 1: 'Sequence Entry'
# Button presses set user input to keywords that initiate individual subunit processes (steps 1 - 8 above)
site = STC.EditSite('CAT', Cas9)
display = STC.DisplayStrings(site)
progress_tracker.set(1)
section_loop_exit.set(False)
enter_sequence_setup()

while True:
    if section_loop_exit.get():                     # This loop prevents deep recursive calls to program_controller
        section_loop_exit.set(False)                # loop_exit is set to true when the program_controller step loop
        program_controller(user_response.get())     # is exited by button press calling a different step
    else:
        root.wait_variable(user_response)
        if user_response.get() in dict_program_control.values():
            program_controller(user_response.get())

root.mainloop()